#---a popup window for choicing files
#
# -type "Save_as | Open (default)"
# -initialdir "<start in dir> | home-dir (default)"
# -initialfile "<your file name proposal> | nothing (default)"
# -multiple "0 (default) | 1 (selecting of multiple files allowed)"
# -filetype "[list "Description" "*.file extension"]"
# -only_filetype "0 add "All files" "*.*" to the -filedir (default) | 1 only types under -filetype"
# -view_hidden_files "0 (default) | 1"
# -view_hidden_dirs "0 (default) | 1"
# -title "<your window title> | "Please choose a file:" (default)"
# -parent "<window which starts FR_fileBox> | . (default)"
# -permission "writable | readable (default)"
# -balloonhelp "0 (default) | 1"

package provide FR_fileBox 0.1

foreach Arg [list FR_general FR_comboBox FR_messageBox FR_popupWindow FR_balloonHelp] \
{
 package require $Arg
}

namespace eval ::FR_fileBox \
{
 namespace export Create
}


#---the content of a directory
proc ::FR_fileBox::DirTreeContent {Path Dir} \
{
 variable VAR
 
 set Content [glob -directory $Dir -types {d} -tails -nocomplain *]
 
 if {$VAR($Path,HiddenDirs)} \
 {
  set Tmp [glob -directory $Dir -types {d hidden} -tails -nocomplain *]
  set Tmp [lrange $Tmp 2 end]
  eval lappend Content $Tmp
 }
 
 set Content [lsort -dictionary $Content]
   
 if { $Dir ne "/" } { set Content [linsert $Content 0 ".."] }
   
 return $Content
}


#---fills the directories into the listbox
proc ::FR_fileBox::DirTreeView {Path Dir} \
{   
 variable VAR
   
 set Content [::FR_fileBox::DirTreeContent $Path $Dir]
    
 .$VAR($Path,dirtree).dirtree delete 0 end
      
 foreach Arg $Content \
 {
  .$VAR($Path,dirtree).dirtree insert end $Arg
 }
}


#---find the file type of the combobox and return it with glob pattern
proc ::FR_fileBox::ShortFind {Arg} \
{
 set ShortFound ""
 #---we are looking for *.(letters or numbers or *)
 set ShortFoundTmp [regexp -all -inline -- {\*.([A-Za-z0-9\*]+)} $Arg]
  
 #---we sort out the search return with the whole line
 for {set i 0} {$i < [llength $ShortFoundTmp]} {incr i 2} \
 {
  lappend ShortFound [lindex $ShortFoundTmp $i]
 }
  
 #---we replace e.g *.wav with *.[Ww][Aa][Vv]
 for {set i 0} {$i < [llength $ShortFound]} {incr i} \
 {
  set NewArg ""
  set Arg [lindex $ShortFound $i]
   
  #---in case of *.* we set it to * to fetch files without tail also (e.g. Readme)
  if {$Arg eq "*.*"} \
  {
   set NewArg "*"
  } else \
  {
   for {set j 0} {$j < [string length $Arg]} {incr j} \
   {
    set Letter [string index $Arg $j]
    
    if {[string is lower $Letter]} \
    {
     set Letter2 [string toupper $Letter]
     set NewLetter "\[${Letter2}${Letter}\]"
     append NewArg $NewLetter 
    } \
    elseif {[string is upper $Letter]} \
    {
     set Letter2 [string tolower $Letter]
     set NewLetter "\[${Letter}${Letter2}\]"
     append NewArg $NewLetter 
    } else \
    {
     append NewArg $Letter
    }
   }
  }
   
  lset ShortFound $i $NewArg
 }
  
 return $ShortFound
} 


#---returns the files in "dir"
proc ::FR_fileBox::FileTreeContent {Path Dir} \
{
 variable VAR
  
 set FileType [::FR_comboBox::cb .$VAR($Path,entry).ftype -read]
  
 set FileType [::FR_fileBox::ShortFind $FileType]
    
 set Content [glob -directory $Dir -types {f} -tails -nocomplain $FileType]
  
 if {$VAR($Path,HiddenFiles)} \
 {
  set HiddenFiles [glob -directory $Dir -types {hidden f} -tails -nocomplain $FileType]
   
  if {[llength $HiddenFiles] > 0} \
  {
   eval lappend Content $HiddenFiles
  }
 }   
     
 set Content [lsort -dictionary $Content]
   
 return $Content
}


#---fills the files into the listbox
proc ::FR_fileBox::FileTreeView {Path Dir} \
{   
 variable VAR
   
 set Content [::FR_fileBox::FileTreeContent $Path $Dir]
  
 .$VAR($Path,files).files delete 0 end  
  
 foreach Arg $Content \
 {
  .$VAR($Path,files).files insert end $Arg
  .$VAR($Path,files).files itemconfigure end -foreground blue
 }
}
 
 
#---fills the choiced file(s) into the file-entry-box
proc ::FR_fileBox::InsertFile {Path Item} \
{
 variable VAR

 .$VAR($Path,entry).entry delete 0 end
 .$VAR($Path,entry).entry insert end $Item
}


#---fills the choiced directory into the dir-entry-box
proc ::FR_fileBox::InsertDir {Path Item} \
{
 variable VAR

 .$VAR($Path,entry).currentdir configure -state normal
  
 .$VAR($Path,entry).currentdir delete 0 end
 .$VAR($Path,entry).currentdir insert end $Item
  
 .$VAR($Path,entry).currentdir configure -state disable
}    
 

#---the reaction to a selection in the dir-listbox
proc ::FR_fileBox::DirTreeSelection {Path Widget} \
{
 variable VAR
 
 set Index [$Widget curselection]
   
 if {$Index eq ""} {return}
   
 set Selected [$Widget get $Index]
   
 if { $Selected ne ".." } \
 {
  set VAR($Path,TmpDir) [file join $VAR($Path,TmpDir) $Selected]
 } else \
 {
  set LastSlash [string last / $VAR($Path,TmpDir)]
     
  #---to avoid the cutting of the / of the root-directory
  if { $LastSlash > 1} {incr LastSlash -1}
     
  set VAR($Path,TmpDir) [string range $VAR($Path,TmpDir) 0 $LastSlash]
 }
    
 ::FR_fileBox::InsertDir $Path $VAR($Path,TmpDir)
    
 ::FR_fileBox::DirTreeView $Path $VAR($Path,TmpDir)
 ::FR_fileBox::FileTreeView $Path $VAR($Path,TmpDir)
}

 
#---the reaction to a single click selection in the file-listbox 
proc ::FR_fileBox::FilesSelection {Path Widget} \
{
 variable VAR
   
 set Selected ""
   
 set Index [$Widget curselection]
 if {$Index eq ""} {return}
   
 foreach i $Index {lappend Selected [$Widget get $i]}
 ::FR_fileBox::InsertFile $Path $Selected
}
 
#---the reaction to a double click selection in the file-listbox
proc ::FR_fileBox::FilesSelectionDouble {Path Widget} \
{
 variable VAR
   
 ::FR_fileBox::FilesSelection $Path $Widget
  
 .$VAR($Path,botton).ok invoke
}
 
 
#---refresh the dir- and the filetree-listboxes
proc ::FR_fileBox::DirTreeRefresh {Path {Dir ""}} \
{
 variable VAR

 if {$Dir ne ""} \
 {
  eval set VAR($Path,TmpDir) $Dir
 }
  
 ::FR_fileBox::InsertDir $Path $VAR($Path,TmpDir)
    
 ::FR_fileBox::DirTreeView $Path $VAR($Path,TmpDir)
 ::FR_fileBox::FileTreeView $Path $VAR($Path,TmpDir)
}

 
#---we check if all demanded file rights are fullfiled
proc ::FR_fileBox::RightsCheck {Path Item} \
{
 variable VAR
  
 if { $VAR($Path,Permission) eq "writable" } \
 { 
  set Cmd [list file writable $Item]       
  set Text [msgcat::mc "write-"]
 } else \
 {
  set Cmd [list file readable $Item]
  set Text [msgcat::mc "read-"]
 }
     
 if { ![eval $Cmd] } \
 {
  ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "You need the %spermission for this file, but you havn't it." $Text]
       
  ::FR_fileBox::DirTreeRefresh $Path $VAR($Path,TmpDir)
       
  return 0
 } else \
 {
  return 1
 }
}
 
 
#---we generate the stricpt-return-value
proc ::FR_fileBox::EntryCheck {Path} \
{
 variable VAR
  
 set TestedFiles ""
 set FileListTmp ""
    
 set FileList [.$VAR($Path,entry).entry get]
  
 if {[llength $FileList] < 1} {return}
  
 foreach Arg $FileList \
 {
  set Tmp [file join $VAR($Path,TmpDir) $Arg]
  # we mask the spaces
  regsub -all {\s} $Tmp {\\ } Tmp
  
  lappend FileListTmp $Tmp
 }

 set FileList $FileListTmp

  
 if {$VAR($Path,Type) eq "Save_as"} \
 {
  foreach Arg $FileList \
  {
   set DirCheck [file dirname $Arg]
    
   if {[file writable $DirCheck]} \
   {
    lappend TestedFiles $Arg
   }
  }
 }
  
  
 if {$VAR($Path,Type) eq "Open"} \
 {
  foreach Arg $FileList \
  {
   if {[file isfile $Arg]} \
   {
    if {[::FR_fileBox::RightsCheck $Path $Arg]} {lappend TestedFiles $Arg}
   }
  }
 }
   
 if { [llength $TestedFiles] > 0 } \
 {
  set VAR($Path,SelectedFiles) $TestedFiles
  focus $VAR($Path,Parent)
  destroy .$Path
  set VAR($Path,Stop) 1
 }
}
 
 
#---we make a new directory (via menu)
proc ::FR_fileBox::NewDir {Path} \
{
 variable VAR
  
 set Dir [::FR_popupWindow::create -parent .$Path -initialinput $VAR($Path,TmpDir) -description [msgcat::mc "Input your new directory:"]]
  
 if {$Dir eq ""} {return 1}
  
 set Dir [file normalize $Dir]
   
 if {[file exists $Dir]} \
 {
  ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "This directory already exists."]
    
  ::FR_fileBox::DirTreeRefresh $Path $Dir
    
  return 1
 }
   
 if { ![catch { file mkdir $Dir } ReturnContent] } \
 {
  ::FR_fileBox::DirTreeRefresh $Path $Dir
 } else \
 {
  ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "I can't create this directory!\nThe reason is:\n\n%s" $ReturnContent]

  ::FR_fileBox::DirTreeRefresh $Path
 }  
}
 

#---with this you can change into a directory directly (via menu)
proc ::FR_fileBox::ChangeDir {Path} \
{
 variable VAR
  
 set Dir [::FR_popupWindow::create -parent .$Path -initialinput $VAR($Path,TmpDir) -description [msgcat::mc "Input your new directory:"]]
  
 if {$Dir eq ""} {return 1}
  
 set Dir [file normalize $Dir]
  
 if {[file isdirectory $Dir]} \
 {
  ::FR_fileBox::DirTreeRefresh $Path $Dir
 } else \
 {
  ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "%s isn't a valid directory" $Dir]
 }
}


#---with this you can delete a directory (via menu)   
proc ::FR_fileBox::DeleteDir {Path} \
{
 variable VAR
   
 set Index [.$VAR($Path,dirtree).dirtree curselection]
  
 if {$Index eq ""} \
 {
  ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "Please select first a directory."]
 } else \
 { 
  set Dir [.$VAR($Path,dirtree).dirtree get $Index]
   
  set Dir [file join $VAR($Path,TmpDir) $Dir]
   
  if { ![catch { file delete -force $Dir } ReturnContent] } \
  {
   ::FR_fileBox::DirTreeRefresh $Path
  } else \
  {
   ::FR_messageBox::Create -type Error -parent .$Path -message [msgcat::mc "I can't delete this directory!\nThe reason is:\n\n%s" $ReturnContent]
  } 
 } 
}


#---the reaction after pressing the cancel button
proc ::FR_fileBox::CancelButton {Path} \
{
 variable VAR
  
 focus $VAR($Path,Parent)
 destroy .$Path
 set VAR($Path,SelectedFiles) ""
 set VAR($Path,Stop) 1
}


#---we save the old content of the entry box, we need this for fileentry_cancel
proc ::FR_fileBox::FileEntrySaveOld {Path} \
{
 variable VAR
  
 set VAR($Path,EntryOldContent) [.$VAR($Path,entry).entry get]
}
 
  
#---procedure to abort the new directory making, set the entry window in the standard state
proc ::FR_fileBox::FileEntryCancel {Path {AutoFocus 1}} \
{
 variable VAR
  
 #---if canceled and anything was written in the newdir_text - we clean this
 ::FR_fileBox::InsertFile $Path $VAR($Path,EntryOldContent)
    
 #---set focus to the directory tree or not (in the focusout case)
 if { $AutoFocus == 1 } { focus  .$VAR($Path,dirtree).dirtree }
}


#---the ballonhelp
proc ::FR_fileBox::BalloonHelp {Path} \
{
 variable VAR
 
 set ::FR_balloonHelp::wl(.$VAR($Path,dirtree).dirtree) [msgcat::mc "With a right-mouse-button-klick you get a menu with some directory functions"]
 set ::FR_balloonHelp::wl(.$VAR($Path,entry).entry) [msgcat::mc "You can here input directly the filename"]

 if {$VAR($Path,BalloonHelp)} {::FR_balloonHelp::create}
}


#---------------------
#---the main procedure
#---------------------

proc ::FR_fileBox::Create {args} \
{
 global env
 variable VAR

 #---first we generate a toplevel name who doesn't exist yet
 set i 1
 while {[winfo exists .filebox_dialog$i]} {incr i}
 set Path "filebox_dialog$i"

 #---some default settings
 set VAR($Path,Type) ""
 set VAR($Path,TmpDir) "$env(HOME)"
 set VAR($Path,IniFile) ""
 set VAR($Path,Multiple) 0
 set VAR($Path,FileType) ""
 set VAR($Path,OnlyFileType) 0
 set VAR($Path,HiddenFiles) 0
 set VAR($Path,HiddenDirs) 0
 set VAR($Path,Parent) .
 set VAR($Path,Permission) "readable"
 set VAR($Path,BalloonHelp) 0
  
 #---we read the arguments
 for {set i 0} {$i < [llength $args]} {incr i 2} \
  {
   set Arg [lindex $args $i]
   switch -glob -- $Arg \
    {
     -type              {set VAR($Path,Type) [lindex $args [expr {$i+1}]]} \
     -initialdir        {set VAR($Path,TmpDir) [lindex $args [expr {$i+1}]]} \
     -initialfile       {set VAR($Path,IniFile) [lindex $args [expr {$i+1}]]} \
     -multiple          {set VAR($Path,Multiple) [lindex $args [expr {$i+1}]]} \
     -filetype          {set VAR($Path,FileType) [lindex $args [expr {$i+1}]]} \
     -only_filetype     {set VAR($Path,OnlyFileType) [lindex $args [expr {$i+1}]]} \
     -view_hidden_files {set VAR($Path,HiddenFiles) [lindex $args [expr {$i+1}]]} \
     -view_hidden_dirs  {set VAR($Path,HiddenDirs) [lindex $args [expr {$i+1}]]} \
     -title             {set VAR($Path,Title) [lindex $args [expr {$i+1}]]} \
     -parent            {set VAR($Path,Parent) [lindex $args [expr {$i+1}]]} \
     -permission        {set VAR($Path,Permission) [lindex $args [expr {$i+1}]]} \
     -balloonhelp       {set VAR($Path,BalloonHelp) [lindex $args [expr {$i+1}]]} \
     --                 {break} \
     *                  {puts "Unknown option: $Arg"; exit 1}
    }
   }


 #---we wait for an input on the end of FR_fileBox
 set VAR($Path,Stop) 0


 #---we generate a new toplevel window
 toplevel .$Path
 
 #---we check the arguments
 #---if the type doesn't set to "Save_as" -> we set it "Open"
 if {$VAR($Path,Type) eq "Save_as"} \
 {
  set VAR($Path,Title) [msgcat::mc "Save as:"]
 } else \
 {
  set VAR($Path,Type) "Open"
  set VAR($Path,Title) [msgcat::mc "Please choose a file:"]
 }

 #---if a wrong directory set -> we set it to HOME
 if {![file isdirectory $VAR($Path,TmpDir)]} \
 {
  set VAR($Path,TmpDir) "$env(HOME)"
 }

 #---if "ONLY_FILE_TYPE" didn't set -> we add "All files" to "FILE_TYP"
 if {!$VAR($Path,OnlyFileType)} \
 {
  lappend VAR($Path,FileType) [msgcat::mc "All files"] "*.*"
 }

 #---if no file type set -> we set it to "All files"
 if {$VAR($Path,FileType) eq ""} \
 {
  set VAR($Path,FileType) [list [msgcat::mc "All files"] "*.*"]
 }

 #---if no parent window set -> we set it to "."
 if {![winfo exists $VAR($Path,Parent)]} \
 {
  set VAR($Path,Parent) "."
 }

 #---if the permission rights don't set to "writeable" -> we set it "readable"
 if {![expr {$VAR($Path,Permission) eq "writeable"}] } \
 { 
  set VAR($Path,Permission) "readable"
 }

 #---we set the selectmode of the file (only for the type "Open")
 if {($VAR($Path,Multiple)) && ($VAR($Path,Type) eq "Open")} \
 {
  set VAR($Path,ListBoxMode) "extended"
 } else \
 {
  set VAR($Path,ListBoxMode) "browse"
 }

 
 #---we set the windowsname and hide it
 wm title .$Path "$VAR($Path,Title)"
 wm withdraw .$Path   


 #---we format the file types for the combobox
 #---DescLength is for the description of the file typ format, ShortLength is for the (glob-)short form
 set DescLength 25
 set ShortLength 15
 set VAR($Path,FileTypeCBox) ""
 
 for {set i 0} {$i < [llength $VAR($Path,FileType)]} {incr i 2} \
 {
  #---we read out the description and the short form
  set j [expr {$i+1}]
  
  set FileTypeDesc [lindex $VAR($Path,FileType) $i]
  set FileTypeDescLength [string length $FileTypeDesc]
  
  set FileTypeShort [lindex $VAR($Path,FileType) $j]
  set FileTypeShortLength [string length $FileTypeShort]
  
  #---we format the description to DescLength (with cutting or filling with spaces)
  if {$FileTypeDescLength > $DescLength} \
  {
   set FileTypeDesc [string range $FileTypeDesc 0 [expr {$DescLength - 2}]]
   append FileTypeDesc "~"
  } else \
  {
   append FileTypeDesc [string repeat " " [expr {$DescLength - $FileTypeDescLength}]]
  }
  
  #---we format the short form to ShortLength (with cutting or filling with %s)
  if {$FileTypeShortLength > $ShortLength} \
  {
   set FileTypeShort [string range $FileTypeShort 0 [expr {$ShortLength - 2}]]
   append FileTypeShort "~"
  } else \
  {
   append FileTypeShort [string repeat " " [expr {$ShortLength - $FileTypeShortLength}]]
  }
  
  #---we put (formated) description and short form together
  append FileTypeDesc $FileTypeShort
  
  #---we replace the old description
  lset VAR($Path,FileType) $i $FileTypeDesc
  
  #---we create a list with the new description for the the combobox
  lappend VAR($Path,FileTypeCBox) $FileTypeDesc
 }
  
  
 #---the frame for the buttons
 frame .$Path.frame2 -relief flat
 frame .$Path.frame3 -relief flat
 labelframe .$Path.frame4 -relief groove -text [msgcat::mc "Directory tree"]
 labelframe .$Path.frame5 -relief groove -text [msgcat::mc "Files"] -foreground blue
  
 pack  .$Path.frame2 -side bottom -fill x
 pack  .$Path.frame3 -side bottom -fill x
 pack  .$Path.frame4 -side left -padx 5 -pady 5 -expand yes -fill both
 pack  .$Path.frame5 -side left -padx 5 -pady 5 -expand yes -fill both
 
 #---We set a variable for each frame - so we can change the order of the frames later easier
 set VAR($Path,botton) "$Path.frame2"
 set VAR($Path,entry) "$Path.frame3"
 set VAR($Path,dirtree) "$Path.frame4"
 set VAR($Path,files) "$Path.frame5"
 
 #---a menu for some dir-operations
 menu .$VAR($Path,dirtree).menu -tearoff 0
 .$VAR($Path,dirtree).menu add command -label [msgcat::mc "Go to home directory"] -command [list ::FR_fileBox::DirTreeRefresh $Path $env(HOME)]
 .$VAR($Path,dirtree).menu add separator
 .$VAR($Path,dirtree).menu add command -label [msgcat::mc "Go to / -directory"] -command [list ::FR_fileBox::DirTreeRefresh $Path /]
 .$VAR($Path,dirtree).menu add separator
 .$VAR($Path,dirtree).menu add command -label [msgcat::mc "Go direct in another directory"] -command [list ::FR_fileBox::ChangeDir $Path]
 .$VAR($Path,dirtree).menu add command -label "" -command {}
 .$VAR($Path,dirtree).menu add command -label [msgcat::mc "Add a new directory"] -command [list ::FR_fileBox::NewDir $Path]
 .$VAR($Path,dirtree).menu add separator
 .$VAR($Path,dirtree).menu add command -label [msgcat::mc "Delete a directory"] -command [list ::FR_fileBox::DeleteDir $Path]

 
 #---the listbox for the directories
 listbox .$VAR($Path,dirtree).dirtree -background white -activestyle none \
 -yscroll ".$VAR($Path,dirtree).yscrollbar set" -xscroll ".$VAR($Path,dirtree).xscrollbar set"
 
 scrollbar .$VAR($Path,dirtree).yscrollbar -orient vertical -command ".$VAR($Path,dirtree).dirtree yview"
 scrollbar .$VAR($Path,dirtree).xscrollbar -orient horizontal -command ".$VAR($Path,dirtree).dirtree xview"
 
 grid .$VAR($Path,dirtree).dirtree -row 0 -column 0 -sticky nsew -padx {5 0} -pady {5 0}
 grid .$VAR($Path,dirtree).yscrollbar -row 0 -column 1 -sticky ns -padx {0 5} -pady {5 0}
 grid .$VAR($Path,dirtree).xscrollbar -row 1 -column 0 -sticky ew -padx {5 0} -pady {0 5}

 grid columnconfigure .$VAR($Path,dirtree) 0 -weight 1
 grid rowconfigure .$VAR($Path,dirtree) 0 -weight 1
 
 bind .$VAR($Path,dirtree).dirtree <Double-Button-1> [list ::FR_fileBox::DirTreeSelection $Path %W]
 bind .$VAR($Path,dirtree).dirtree <Button-3> [list tk_popup .$VAR($Path,dirtree).menu %X %Y]


 #---the listbox for the files
 listbox .$VAR($Path,files).files -background white -activestyle none -selectmode $VAR($Path,ListBoxMode) \
 -yscroll ".$VAR($Path,files).yscrollbar set" -xscroll ".$VAR($Path,files).xscrollbar set"
 
 scrollbar .$VAR($Path,files).yscrollbar -orient vertical -command ".$VAR($Path,files).files yview"
 scrollbar .$VAR($Path,files).xscrollbar -orient horizontal -command ".$VAR($Path,files).files xview"
 
 grid .$VAR($Path,files).files -row 0 -column 0 -sticky nsew -padx {5 0} -pady {5 0}
 grid .$VAR($Path,files).yscrollbar -row 0 -column 1 -sticky ns -padx {0 5} -pady {5 0}
 grid .$VAR($Path,files).xscrollbar -row 1 -column 0 -sticky ew -padx {5 0} -pady {0 5}

 grid columnconfigure .$VAR($Path,files) 0 -weight 1
 grid rowconfigure .$VAR($Path,files) 0 -weight 1
 
 bind .$VAR($Path,files).files <<ListboxSelect>> [list ::FR_fileBox::FilesSelection $Path %W]
 bind .$VAR($Path,files).files <Double-Button-1> [list ::FR_fileBox::FilesSelectionDouble $Path %W]
   
 #---The ok and chancel buttons
 set ButtonText [::FR_general::ButtonsWithSameWidth [list [msgcat::mc "Ok"] [msgcat::mc "Cancel"]]]
 
 button .$VAR($Path,botton).ok -text [lindex $ButtonText 0] -state normal -width [lindex $ButtonText 2] -command [list ::FR_fileBox::EntryCheck $Path]
  
 button .$VAR($Path,botton).cancel -text [lindex $ButtonText 1] -state normal -width [lindex $ButtonText 2] -command [list ::FR_fileBox::CancelButton $Path]
  
 pack .$VAR($Path,botton).ok -side left -padx 5 -pady 5 -fill x -expand yes
 pack .$VAR($Path,botton).cancel -side left -padx 5 -pady 5 -fill x -expand yes
 
 bind .$Path <KeyPress-Escape> [list .$VAR($Path,botton).cancel invoke]
 bind .$Path <KeyPress-Return> [list .$VAR($Path,botton).ok invoke]
 
 
 #---View the current directory
 label .$VAR($Path,entry).currentdir_label -text "Current directory:"
 entry .$VAR($Path,entry).currentdir -state disable -disabledforeground black -disabledbackground white -xscrollcommand ".$VAR($Path,entry).currentdir_xsbar set"
 scrollbar .$VAR($Path,entry).currentdir_xsbar -orient horizontal -command ".$VAR($Path,entry).currentdir xview"
  
 label .$VAR($Path,entry).entry_label -text "File name:"
 entry .$VAR($Path,entry).entry -state normal -disabledforeground black -background white -xscrollcommand ".$VAR($Path,entry).entry_xsbar set"
 scrollbar .$VAR($Path,entry).entry_xsbar -orient horizontal -command ".$VAR($Path,entry).entry xview"
 
 bind .$VAR($Path,entry).entry <FocusIn> [list ::FR_fileBox::FileEntrySaveOld $Path]
 bind .$VAR($Path,entry).entry <KeyPress-Return> [list ::FR_fileBox::EntryCheck $Path]
 bind .$VAR($Path,entry).entry <KeyPress-Escape> [list ::FR_fileBox::FileEntryCancel $Path]

 grid .$VAR($Path,entry).currentdir_label -row 0 -column 0 -sticky w -padx {5 0}
 grid .$VAR($Path,entry).currentdir -row 0 -column 1 -sticky ew -padx {5 0}
 grid .$VAR($Path,entry).currentdir_xsbar -row 0 -column 2 -padx {0 5}
      
 grid .$VAR($Path,entry).entry_label -row 1 -column 0 -sticky w -padx {5 0}
 grid .$VAR($Path,entry).entry -row 1 -column 1 -sticky ew -padx {5 0}
 grid .$VAR($Path,entry).entry_xsbar -row 1 -column 2 -padx {0 5}
 
 
 #---The file type combobox
 label .$VAR($Path,entry).ftype_label -text "Files of type:"
 ::FR_comboBox::cb .$VAR($Path,entry).ftype -entry_width auto -listbox_content $VAR($Path,FileTypeCBox) -parent .$Path -command [list ::FR_fileBox::DirTreeRefresh $Path]
 
 grid .$VAR($Path,entry).ftype_label -row 2 -column 0 -sticky w -pady {0 5} -padx 5
 grid .$VAR($Path,entry).ftype -row 2 -column 1 -columnspan 2 -sticky ew -pady {0 5} -padx 5
 
 grid columnconfigure .$VAR($Path,entry) 1 -weight 1
 
 
 #---at the first start: we fill the dir-listbox and possibly the file-entry
 ::FR_fileBox::DirTreeRefresh $Path $VAR($Path,TmpDir)

 if {$VAR($Path,IniFile) ne ""} \
 {
  ::FR_fileBox::InsertFile $Path $VAR($Path,IniFile)
 }

 ::FR_fileBox::BalloonHelp $Path
 
 #---we place the window
 ::FR_general::CenterWindow .$Path $VAR($Path,Parent)
 wm deiconify .$Path
 focus .$Path

 wm protocol .$Path WM_DELETE_WINDOW [list .$VAR($Path,botton).cancel invoke]
 
 #---We stop the procedure and wait until the return value is ready

 vwait ::FR_fileBox::VAR($Path,Stop)
 return "$VAR($Path,SelectedFiles)"
}

