package provide FR_search 0.2

# Search in files function
#
# return value is the VAR(SearchList) with the elements:
# "filename" "row of the line" "content of the line"
#
# in case of an error during open the file: "row of the line"
# is set to zero; "content of the line" has the error message
#
# Possible arguments and its default values
# -----------------------------------------
# -startdir    "directory where the search starts | [pwd] (default)"
# -pattern     "the search pattern | "" (default)"
# -filetype    "your file type for the search | "*.tcl" (default)"
# -case        "0 for no case sensitivity | 1 for case sensitivity (default)"
# -hiddendir   "1 for including hidden dirs | 0 (default)"
# -hiddenfile  "1 for including hidden files | 0 (default)"
# -returnsl    "1 for return the search results | 0 (default) search results are stored in ::FR_search::VAR(SearchList)" 

namespace eval ::FR_search \
{
 namespace export Create
}


#---we open the file and search the pattern in each line
#---the proc stores its results in VAR(SearchList)
proc ::FR_search::SearchPattern {File} \
{
 variable VAR
 
 if {[catch {set Fid [open $File r]} ReturnValue] != 0} \
 {
  lappend VAR(SearchList) $File 0 $ReturnValue
  return 1
 }
 
 set Counter 0
 
 while {[eof $Fid] == 0} \
 {
  set Line [gets $Fid]
  
  incr Counter
  
  if {$VAR(Case)} \
  {
    set Result [string match $VAR(Pattern) $Line]
  } else \
  {
    set Result [string match -nocase $VAR(Pattern) $Line]
  }
 
  if {$Result == 1} \
  {
    lappend VAR(SearchList) $File $Counter $Line 
  } 
 }
  
 close $Fid
 
 return 0
}


#---we find all subdirectories and start there a new directory search;
#---with the files which are in each directory we start the content search
proc ::FR_search::FileList {StartDir} \
{
 variable VAR

 # we go recursivly through all directories
 set DirList [glob -nocomplain -type {d x} -directory $StartDir *]
 
 if {$VAR(HiddenDirs)} \
 {
  set DirListHiddenDirs [glob -nocomplain -type {d hidden} -directory $StartDir *]
  set DirListHiddenDirs [lrange $DirListHiddenDirs 2 end]
  
  eval lappend DirList $DirListHiddenDirs
  unset DirListHiddenDirs
 }
 
 if {[llength $DirList] > 0} \
 {
  foreach Arg $DirList \
  {
   ::FR_search::FileList $Arg
  }
 }
 
 # now we create a list with all files which matching FILETYPE
 eval lappend VAR(FileList) [glob -nocomplain -type {f} -directory $StartDir $VAR(FileType)]
 
 if {$VAR(HiddenFile)} \
 {
  eval lappend VAR(FileList) [glob -nocomplain -type {f hidden} -directory $StartDir $VAR(FileType)]
 }
  
 return 0
}


#---the start proc for search in files
proc ::FR_search::Create {args} \
{
 variable VAR
 
 # we parse the arguments
 set VAR(StartDir) [pwd]
 set VAR(Pattern) ""
 set VAR(FileType) "*.tcl"
 set VAR(Case) 1
 set VAR(HiddenDirs) 0
 set VAR(HiddenFile) 0
 set VAR(ReturnSearchList) 0
 
 for {set i 0} {$i < [llength $args]} {incr i 2} \
 {
  set arg [lindex $args $i]
  switch -glob -- $arg \
  {
   -startdir    {set VAR(StartDir) [lindex $args [expr {$i+1}]]} \
   -pattern     {set VAR(Pattern) [lindex $args [expr {$i+1}]]} \
   -filetype    {set VAR(FileType) [lindex $args [expr {$i+1}]]} \
   -case        {set VAR(Case) [lindex $args [expr {$i+1}]]} \
   -hiddendir   {set VAR(HiddenDirs) [lindex $args [expr {$i+1}]]} \
   -hiddenfile  {set VAR(HiddenFile) [lindex $args [expr {$i+1}]]} \
   -returnsl    {set VAR(ReturnSearchList) [lindex $args [expr {$i+1}]]} \
   --           {break} \
   *            {puts "Unknown option: $arg"; exit 1}
   }
  }
 
 # now we reset the seach list
 set VAR(SearchList) ""
 # and the filelist
 set VAR(FileList) ""
 
 
 if {$VAR(Pattern) eq ""} {return $VAR(SearchList)}
 
 # we map all special token from string match
 set VAR(Pattern) [string map {{?} {\?} {*} {\*} {[} {\[} {]} {\]}} $VAR(Pattern)]
 set VAR(Pattern) "*${VAR(Pattern)}*"
 
 # now we start the creation of VAR(FileList) which contains all files where Pattern is matched 
 ::FR_search::FileList $VAR(StartDir)
 
 if {[llength $VAR(FileList)] > 0} \
 {
  foreach Arg $VAR(FileList) \
  {
   ::FR_search::SearchPattern $Arg
  }
 }
 
 # ::FR_search::SearchPattern has generated VAR(SearchList)
 if {$VAR(ReturnSearchList)} \
 {
  return $VAR(SearchList)
 } else \
 {
  return 0
 }
}
