#!/bin/sh
#\
exec tclsh "$0" "$@"

# Written by Falk Richter
#
# tk_edit (for a short description please see the About-information) 
#
# This software stands under the BSD LICENSE and comes with 
# NO warranty of any kind. Use at your own risk.
# Report bugs or suggestions to: please see contact in the about-box
# ===============================================================================

namespace eval ::tk_edit \
{
  namespace export Main
}


# we load the msgcat package in a global scope  
package require msgcat

# we check if a language is set from the operation system
if {![info exist env(LANG)]} {set env(LANG) "en"}
msgcat::mclocale $env(LANG)
msgcat::mcload [file join [pwd] msgs]


#---the mother of all procedures :o)
proc ::tk_edit::Main {} \
{
  variable CONFIG; variable VAR
 
  ::tk_edit::Startup
 
  wm protocol . WM_DELETE_WINDOW ::tk_edit::Exit

  ::tk_edit::Frames 
 
  ::tk_edit::Gui
 
  ::tk_edit::Menu
 
  #::tk_edit::BalloonHelp
 
  ::tk_edit::FirstStart
 
  wm geometry . $CONFIG(Geometry)
  wm deiconify .
  update
}


#---here we load all need packages, set the defaults, read and parse the config file
proc ::tk_edit::Startup {} \
{
  variable VAR; variable ABOUT;
  variable CONFIGFILE; variable DEFAULTS; variable CONFIG;
  global env auto_path tcl_platform
 
  package require Tk
  wm withdraw .
 
  set VAR(NameSpace) [namespace current]
  set VAR(StartDir) [pwd]
  
  # we check the existens of lib directory   
  set VAR(LibDir) [file join $VAR(StartDir) lib]
 
  if { [file exists $VAR(LibDir)] !="1" } \
  {
    tk_messageBox -message [msgcat::mc "I can't find the program libraries!\n\nPlease install the program again."]\
    -title [msgcat::mc "Where are my libraries?"] -icon error -type ok
    destroy .
    exit
  } else \
  {
    lappend auto_path $VAR(LibDir)
  }
 
  # we load all needed packages
  ::tk_edit::Packageload [list FR_notebook FR_entry FR_messageBox FR_fileBox FR_config FR_general]
  
  # the informations for the about boxes and the window border
  set InfoFile [file join $VAR(StartDir) ".info.tcl"]
 
  if {[file exists $InfoFile]} \
  {
    source $InfoFile
  } else \
  {
    tk_messageBox -message [msgcat::mc "I can't find the file:\n\n\"%s\"\n\nPlease install the program again." $InfoFile]\
    -title [msgcat::mc "File not found"] -icon error -type ok
    destroy .
    exit
  }
 
  # some settings for the toplevel windows
  set VAR(Parent) .
  wm title $VAR(Parent) "$ABOUT(Name) version $ABOUT(Version) from $ABOUT(Date)"
  wm minsize $VAR(Parent) 800 400
 
  # the default settings

  #set DEFAULTS(BALLOON_HELP) "1"
  set DEFAULTS(Color) "sandyellow"
  set DEFAULTS(Geometry) "+20+20"
  set DEFAULTS(ShowProcWin) 1
  set DEFAULTS(SortProcWin) 1
  set DEFAULTS(FileDirectroy) [pwd]
  # the search in files directory
  set DEFAULTS(SifDirectroy) $DEFAULTS(FileDirectroy)
  set DEFAULTS(SifHiddenDirs) 0
  set DEFAULTS(SifHiddenFiles) 0
  set DEFAULTS(SifOnlyTclFiles) 1
  set DEFAULTS(SifCaseSens) 1
   
  set DEFAULTS(FileTypes) [list [msgcat::mc "Tcl-files"] "*.tcl" [msgcat::mc "Text-files"] "*.txt"]
  set DEFAULTS(CurrentFiles) ""
 
  set DEFAULTS(TextFont) [list "Fixed 12"]
  set DEFAULTS(TextForeGround) white
  set DEFAULTS(TextBackGround) black
  set DEFAULTS(TextIndent) 1
  set DEFAULTS(Backup) 1
  set DEFAULTS(CaseSens) 0
 
  # the borderwidth for the notebook widget
  set VAR(BorderWidth) 1
 
  # we read and parse the configuration file
  set ConfigFileDir [file join $env(HOME) .${ABOUT(Name)}]
 
  if {![file exists $ConfigFileDir]} \
  {
    catch {file mkdir $ConfigFileDir}
  }
 
  set CONFIGFILE [file join $ConfigFileDir .${ABOUT(Name)}_conf.tcl]
 
  ::FR_config::ReadConfig $VAR(NameSpace)

  # we check the config directories and set to the default values if the directory doesn't exist
  foreach Arg {FileDirectroy SifDirectroy} \
  {
    if { ![file isdirectory $CONFIG($Arg)] } \
    {
      set CONFIG($Arg) $DEFAULTS($Arg)
    } 
  }
 
  # we define the "look" (colours)
  set OptionsFile [file join $VAR(StartDir) .tk_options.tcl]
 
  if {[file exists $OptionsFile]} \
  {
    source $OptionsFile
    database_options $VAR(NameSpace)
  }
 
  # we load the ctext source
  set CTextSource [file join $VAR(StartDir) ctext.tcl]

  if {[file exists $CTextSource]} {namespace inscope :: source $CTextSource}
}


#---the package loader
#---we load the needed package only one time
proc ::tk_edit::Packageload {Packages} \
{
  variable VAR

  foreach Arg $Packages \
  {
    if {![info exists VAR($Arg)]} \
    {
      if { [catch {package require $Arg} output] !="0" } \
      {
        tk_messageBox -message [msgcat::mc "I can't load this part from the program library:\n\n%s\n\nPlease install the program again." $Arg]\
        -title [msgcat::mc "Parts of the program couldn't be loaded"] -icon error -type ok
        destroy .
        exit
      } else \
      {
        set VAR($Arg) 1
      }
    }
  }
}


#---we start a external proc but load this file only ones
proc ::tk_edit::ProcStart {Procedure File} \
{
  variable VAR
 
  set Tmp [string trimleft $Procedure "::tk_edit::"]
  if {[info procs $Tmp] eq ""} \
  {
    source [file join $VAR(StartDir) tcl $File]
  }
 
  set ReturnValue [eval $Procedure]
 
  return $ReturnValue
}


#---we save the configuration and finish the program
proc ::tk_edit::Exit {} \
{
  variable CONFIG; variable VAR
 
  set CONFIG(CurrentFiles) ""
 
  foreach Arg [::FR_notebook::GetAllTabs .$VAR(nb)] \
  {  
    ::FR_notebook::Raise .$VAR(nb) -tab $Arg
  
    lappend CONFIG(CurrentFiles) $Arg
  
    ::tk_edit::CloseFile
  }
 
  ::FR_config::SaveConfig $VAR(NameSpace)
 
  catch {puts [msgcat::mc "Goodby!"]}
  destroy .
  exit
}


#---We create the frames and set a variable for each frame - 
#---so we can change the order of the frames later
proc ::tk_edit::Frames {} \
{
  variable VAR
 
  frame .frame1 -relief flat
  frame .frame2 -relief flat
  frame .frame3 -relief raised -borderwidth $VAR(BorderWidth)
  frame .frame4 -relief flat
  
  pack .frame1 -side top -fill x -padx 5
  pack .frame4 -side bottom -fill x -padx 5
  pack .frame2 -side left -fill both -expand yes -padx 5 -pady 5
   
  set VAR(menuframe) "frame1"
  set VAR(mainframe) "frame2"
  set VAR(procframe) "frame3"
  set VAR(bottomframe) "frame4"
}  


#---delete the selected text first before past
proc ::tk_edit::TextPast {} \
{
  variable VAR
 
  set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
  if {$TextFrame eq ""} {return}
  append TextFrame ".t"
 
  set SelPos [.$TextFrame tag ranges sel]

  if {$SelPos ne ""} \
  {
    eval .$TextFrame delete $SelPos
  }
 
  tk_textPaste .$TextFrame
}


#---the text cut function
proc ::tk_edit::TextCut {} \
{
  variable VAR
 
  set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
  if {$TextFrame eq ""} {return}
  append TextFrame ".t"
 
  tk_textCut .$TextFrame
}


#---the text copy function
proc ::tk_edit::TextCopy {} \
{
  variable VAR
 
  set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
  if {$TextFrame eq ""} {return}
  append TextFrame ".t"
 
  tk_textCopy .$TextFrame
}


#---this places the Insert cursor in the text widget
proc ::tk_edit::SetCursor {TextWidget Pos Search} \
{
  variable VAR
 
  # Search comes from the search in file function
  if {$Search} \
  {
    # we search again in this line to...
    set Found [$TextWidget search -count Cnt -- $VAR(Search) $Pos "$Pos lineend"]
  
    # ...select the search result
    if {$Found ne ""} \
    {
      $TextWidget tag remove sel 1.0 end
      $TextWidget tag add sel $Found "$Found + $Cnt chars"
      set Pos "$Found wordend"
    }
  }
  
  $TextWidget mark set insert $Pos
  $TextWidget see insert
  
  update idletasks
  return
}


#---open each file
proc ::tk_edit::OpenFile {{Files {}}} \
{
  variable CONFIG
 
  if {![llength $Files]} \
  {
    set Files [::FR_fileBox::Create -initialdir $CONFIG(FileDirectroy) -view_hidden_files 1 -view_hidden_dirs 1 -filetype $CONFIG(FileTypes)]
  }
 
  if {![llength $Files]} {return}
 
  foreach Arg $Files \
  {
    ::tk_edit::OpenFileSub $Arg "1.0"
  }
  ::tk_edit::FindProcs
}


#---this creates a new textwindow and opens the wished file in it
proc ::tk_edit::OpenFileSub {File Pos {Search {0}}} \
{
  variable CONFIG; variable VAR
 
  if {![file exists $File]} \
  {
    return
  } else \
  {
    set File [file normalize $File]
  }

  # if there already an empty new.tcl we delete this
  set New [file normalize [file join ~ [msgcat::mc new].tcl]]
  if {[lsearch [::FR_notebook::GetAllTabs .$VAR(nb)] $New] != -1} \
  {
    # we check if the empty file is blank yet
    set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $New]
    if {![string length [.${TextFrame}.t get 1.0 end-1c]]} \
    {
      ::FR_notebook::DeleteTab .$VAR(nb) -tab $New
    }
  }
 
  # if a window with this file name already exists we raise it in the forground
  if {[lsearch [::FR_notebook::GetAllTabs .$VAR(nb)] $File] != -1} \
  {
    ::FR_notebook::Raise .$VAR(nb) -tab $File
    set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $File]
    ::tk_edit::SetCursor .${TextFrame}.t $Pos $Search
    return
  }
  
  set CONFIG(FileDirectroy) [file dirname $File]
  set Fid [open $File r]
  set Data [read $Fid]
  close $Fid
  
  set TextWidget [::tk_edit::NewFile $File]
 
  focus -force .$TextWidget
  .$TextWidget insert end $Data
  ::tk_edit::SetCursor .$TextWidget $Pos $Search
}


#---highlighting function
proc ::tk_edit::Highlighting {Window} \
{
  ::ctext::addHighlightClass $Window widgets purple [list bell bind bindtags bitmap button canvas checkbutton clipboard console destroy entry event focus font frame grab grid image label labelframe listbox lower menu menubutton message option pack\
  panedwindow photo place radiobutton raise scale scrollbar selection send spinbox text tk tk_bisque tk_chooseColor tk_chooseDirectory tk_dialog tk_focusFollowsMouse tk_focusNext tk_focusPrev tk_getOpenFile tk_getSaveFile tk_menuSetFocus\
  tk_messageBox tk_optionMenu tk_popup tk_setPalette tk_textCopy tk_textCut tk_textPaste tkerror tkwait toplevel ttk::button ttk::checkbutton ttk::combobox ttk::entry ttk::frame ttk::image ttk::intro ttk::label ttk::labelframe ttk::menubutton\
  ttk::notebook ttk::panedwindow ttk::progressbar ttk::radiobutton ttk::scrollbar ttk::separator ttk::sizegrip ttk::style ttk::treeview ttk::widget winfo wm]
  
  ::ctext::addHighlightClass $Window commands purple [list after append apply array auto_execok auto_import auto_load auto_mkindex auto_mkindex_old auto_qualify auto_reset bgerror binary catch cd chan clock close concat dde dict encoding eof error\
  eval exec exit expr fblocked fconfigure fcopy file fileevent flush format gets glob history incr info interp join lappend lassign lindex linsert list llength load lrange lrepeat lreplace lreverse lset lsort memory namespace open\
  package parray pid pkg_mkIndex puts pwd read regexp registry regsub rename return scan seek set socket split string subst tcl_endOfWord tcl_findLibrary tcl_startOfNextWord tcl_startOfPreviousWord tcl_wordBreakAfter tcl_wordBreakBefore\
  tell time trace unknown unload unset update uplevel upvar vwait]
 
  ::ctext::addHighlightClass $Window flags orange [list -text -command -yscrollcommand -xscrollcommand -background -foreground -fg -bg -highlightbackground -y -x -highlightcolor -relief -width  -height -wrap -font -fill -side -outline -style -insertwidth\
  -textvariable -activebackground -activeforeground -insertbackground -anchor -orient -troughcolor -nonewline -expand -type -message -title -offset -in -after -yscroll -xscroll -forward -count -exact -padx -ipadx -filetypes -all -from -to -label\
  -value -variable -regexp -backwards -forwards -bd -pady -ipady -state -row -column -cursor -highlightcolors -linemap -menu -tearoff -displayof -cursor -underline -tags -tag -weigth -sticky -icon]

  ::ctext::addHighlightClass $Window procs red [list proc]
 
  ::ctext::addHighlightClass $Window stackControl gold [list while for foreach if then elseif else switch break continue]
  ::ctext::addHighlightClassWithOnlyCharStart $Window vars mediumspringgreen "\$"
  ::ctext::addHighlightClass $Window variable_funcs gold [list global variable]
  ::ctext::addHighlightClassForSpecialChars $Window brackets green {[]{}}
  ::ctext::addHighlightClassForRegexp $Window paths lightblue {\.[a-zA-Z0-9\_\-]+}
  ::ctext::addHighlightClassForRegexp $Window comments khaki {#[^\n\r]*}
}


#---generate a new textwidget; if isn't overtaken a file we open the textwidget with new.tcl
proc ::tk_edit::NewFile {{File {}}} \
{
  variable CONFIG; variable VAR
 
  if {![string length $File]} \
  {
    set File [file join ~ [msgcat::mc new].tcl]
  }

  set File [file normalize $File]
 
  # if this file already open -> we raise it in the foreground
  if {[lsearch [::FR_notebook::GetAllTabs .$VAR(nb)] $File] != -1} \
  {
    ::FR_notebook::Raise .$VAR(nb) -tab $File
    return 0
  }
 
  set Paths [::FR_notebook::AddTab .$VAR(nb) -tab $File -text [file tail $File] -extreturn 1]
  set TextFrame [lindex $Paths 0]
  set ButtonPath  [lindex $Paths 1]
 
  bind .$ButtonPath <ButtonPress-3> [list ::tk_edit::NoteBookMenu $File %X %Y]
 
  set Highlight 1

  if {![string match -nocase ".tcl" [file extension $File]]} \
  {
    set Answer [::FR_messageBox::Create -type "Info" -button [list [msgcat::mc "Yes"] [msgcat::mc "No"]] -title [msgcat::mc "Should we highlight this file?"] -message [msgcat::mc "The choiced file hasn't the extension \"tcl\".\n\nShould we highligth it nevertheless?"]]
  
    if {!$Answer} {set Highlight 0}
  }  

  # we create a new text widget
  ctext .${TextFrame}.t -font $CONFIG(TextFont) -undo 1 -linemap 1 -highlight $Highlight -bg $CONFIG(TextBackGround) -fg $CONFIG(TextForeGround) -insertbackground yellow -selectbackground white -wrap none -yscrollcommand ".${TextFrame}.v set" -xscrollcommand ".${TextFrame}.h set"
  scrollbar .${TextFrame}.v -orient vertical -highlightthickness 0 -command ".${TextFrame}.t yview"
  scrollbar .${TextFrame}.h -orient horizontal -highlightthickness 0 -command ".${TextFrame}.t xview"
       
  grid .${TextFrame}.t -row 0 -column 0 -sticky nsew
  grid .${TextFrame}.v -row 0 -column 1 -sticky ns
  grid .${TextFrame}.h -row 1 -column 0 -sticky ew
  grid rowconfigure .${TextFrame} 0 -weight 1
  grid columnconfigure .${TextFrame} 0 -weight 1
 
  bind .${TextFrame}.t <Return> {::tk_edit::HandleEnterKey %W; break}
 
  ::tk_edit::Highlighting .${TextFrame}.t 

  ::FR_notebook::Raise .$VAR(nb) -tab $File
 
  focus -force .${TextFrame}.t
  .${TextFrame}.t mark set insert 1.0
  return ${TextFrame}.t
}


#---we save the file which corresponses to the pressed notebook button
proc ::tk_edit::NoteBookMenu {File X Y} \
{
  variable VAR
  
  set VAR(NbButtonFile) $File
  
  tk_popup .$VAR(mainframe).menu_notebook $X $Y
}


#---the save function
proc ::tk_edit::Save {} \
{
 variable VAR
 
 set File [::FR_notebook::CurrentTab .$VAR(nb)]
 
 if {[string match $File [file normalize [file join ~ [msgcat::mc new].tcl]]]} \
 {
  ::tk_edit::SaveAs
 } else \
 {
  ::tk_edit::WriteFile $File
 }
}


#---the save as function
proc ::tk_edit::SaveAs {{File {}}} \
{
 variable CONFIG; variable VAR
 
 set File [::FR_fileBox::Create -type "Save_as" -initialdir $CONFIG(FileDirectroy) -initialfile [file tail $File] -view_hidden_files 1 -view_hidden_dirs 1 -permission writeable -filetype $CONFIG(FileTypes)]
 
 if {$File eq ""} {return 0}

 ::tk_edit::WriteFile $File
 
 # we raised up the new named file
 ::FR_notebook::DeleteTab .$VAR(nb) -tab [::FR_notebook::CurrentTab .$VAR(nb)]
 ::tk_edit::OpenFile $File
}


#---this checks if the file writeable and write out the file
proc ::tk_edit::WriteFile {File} \
{ 
 variable CONFIG; variable VAR
 
 # we make only a backup when the file already exists
 if {$CONFIG(Backup) && [file exists $File]} \
 {
  set BakFile [file tail $File]
  set BakFile [file rootname $BakFile]
  append BakFile ".bak"
 
  set BakFile [file join [file dirname $File] $BakFile]
 
  file copy -force $File $BakFile
 }
 
 set ReturnValue [catch {set Fid [open $File w]} ReturnContent]
  
 if {$ReturnValue} \
 {
  ::FR_messageBox::Create -type "Error" -title [msgcat::mc "Error during saving"] -message [msgcat::mc "I can not save this file.\n\nThe reason is:\n%s" $ReturnContent]
   
  ::tk_edit::SaveAs $File
 } else \
 {
  set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
  append TextFrame ".t"
   
  puts -nonewline $Fid [.$TextFrame get 1.0 end-1c]
  close $Fid
 }
}


#---compares the saved file with the current content of the text-widget
proc ::tk_edit::Compare {Content File} \
{
 if {[file exists $File] && [file readable $File]} \
 {
  set Fid [open $File r]
  set Data [read $Fid]
  close $Fid
 } else \
 {
  set Data {}
 }
  
 if {[list $Content] ne [list $Data]} \
 {
  return 1
 } else \
 {
  return 0
 }
}


#---this checks if the current file is saved and closes it
proc ::tk_edit::CloseFile {{File {}}} \
{
 variable VAR
 
 set CountNotebookTabs [llength [::FR_notebook::GetAllTabs .$VAR(nb)]]
 
 if {$CountNotebookTabs == 0} {return}
 
 # we call CloseFile from the NB-buttons with a file name
 if {$File eq ""} \
 {
   set File [::FR_notebook::CurrentTab .$VAR(nb)]
 }
 
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $File]
 append TextFrame ".t"
 set Data [.$TextFrame get 1.0 end-1c]
 
 if {[::tk_edit::Compare $Data $File]} \
 {
  set Answer [::FR_messageBox::Create -type "Info" -button [list [msgcat::mc "ok"] [msgcat::mc "no"]] -title [msgcat::mc "Warning - not yet saved"] -message [msgcat::mc "The current file:\n\n%s\n\nis not yet saved!\n\nDo you want to do it?" [file tail $File]]]

  if {$Answer} \
  {
   ::tk_edit::Save
  }
 }
 
 ::FR_notebook::DeleteTab .$VAR(nb) -tab $File
 
 # if we close the last notebook tab we delete the proc window also
 if {$CountNotebookTabs == 1} \
 {
    .$VAR(procframe).procs configure -state normal
    .$VAR(procframe).procs delete 1.0 end
    .$VAR(procframe).procs configure -state disable
 }
}


#---the redo function
proc ::tk_edit::ReDo {} \
{
 variable VAR
 
 set File [::FR_notebook::CurrentTab .$VAR(nb)]
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $File]
 append TextFrame ".t"
 catch ".$TextFrame edit redo"
}


#---the undo function
proc ::tk_edit::UnDo {} \
{
 variable VAR
 
 set File [::FR_notebook::CurrentTab .$VAR(nb)]
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $File]
 append TextFrame ".t"
 catch ".$TextFrame edit undo"
}


#---the select all function
proc ::tk_edit::SelectAll {} \
{
 variable VAR
 
 set File [::FR_notebook::CurrentTab .$VAR(nb)]
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $File]
 append TextFrame ".t"
 .$TextFrame tag add sel 1.0 end-1c
}


#---the ident function from Eric Kemp-Benedict
proc ::tk_edit::HandleEnterKey {Window} \
{
  variable CONFIG
 
  # First, delete the current selection
  if {[$Window tag nextrange sel 1.0] != ""} \
  {
    $Window delete sel.first sel.last
  }

  # Find the whitespace at the start of the current line
  set LineStart [$Window index "insert linestart"]
  set InsertPos [$Window index "insert"]
 
  set WhiteSpaces ""
 
  if {$CONFIG(TextIndent)} \
  {
    regexp {^([\s\t]*)} [$Window get $LineStart $InsertPos] All WhiteSpaces
  }
 
  #puts "WS:${WhiteSpaces}STOP"
  # Create a newline, insert whitespace or tabs
  $Window insert [$Window index insert] "\n$WhiteSpaces"
  # If necessary, scroll the view so cursor is visible
  $Window see [$Window index insert]
}


#---the Comment / Uncomment function
proc ::tk_edit::Comment {{Uncomment 0}} \
{
 variable VAR
 
 # first we look which window is active
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 append TextFrame ".t"
 
 # if isn't anything selected we return
 if {[.$TextFrame tag nextrange sel 1.0] eq ""} then {return}
 
 # we fetch the selection
 set SelStart [.$TextFrame index "sel.first linestart"]
 set SelEnd [.$TextFrame index "sel.last linestart"]
 
 # we separate the lines
 set StartLine [lindex [split $SelStart "."] 0]
 set EndLine [lindex [split $SelEnd "."] 0]
 
 # for each line
 for {set i $StartLine} {$i <= $EndLine} {incr i} \
 { 
  if {$Uncomment} \
  {
   set Line [.$TextFrame get ${i}.0 ${i}.end]
   
   # we are looking for # (which could have spaces or tabs in its front)
   if {[regexp {^([\s\t]*)[#]+} $Line All WhiteSpaces Cross]} \
   {
    # we delete the #
    set Offset [string length $WhiteSpaces]
    set EndCross [expr {$Offset + 1}]
    .$TextFrame delete ${i}.$Offset ${i}.$EndCross
   }
  } else \
  {
   .$TextFrame insert ${i}.0 "#"
  }
 }
 
 # we remove the selection
 .$TextFrame tag remove sel $SelStart $SelEnd
 
 # now we set the cursor at the end of the previous selected block
 incr i -1
 .$TextFrame mark set insert ${i}.end
}


#---the refresh function
proc ::tk_edit::RefreshHighlighting {} \
{
 variable VAR
 
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 append TextFrame ".t"
 
 .$TextFrame configure -highlight 1
 .$TextFrame highlight 1.0 end
}


#---the search function
proc ::tk_edit::FindText {SearchText {LoopCount 0}} \
{
 variable VAR; variable CONFIG
 
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 append TextFrame ".t"
 
 .$VAR(bottomframe).replace_button configure -state disable
 
 if {$SearchText eq ""} {return 0}

 if {$CONFIG(CaseSens)} \
 {
  set Found [.$TextFrame search -- $SearchText [.$TextFrame index insert] end]
 } else \
 {
  set Found [.$TextFrame search -nocase -- $SearchText [.$TextFrame index insert] end]
 }
 
 if {$Found ne ""} \
 {
  .$TextFrame tag remove sel 1.0 end
  .$TextFrame tag add sel $Found "$Found + [string length $SearchText] c"
  .$TextFrame mark set insert "$Found wordend"
  .$TextFrame see "$Found"
  
  .$VAR(bottomframe).replace_button configure -state normal
 } else \
 {
  if {!$LoopCount} \
  {
   set Answer [::FR_messageBox::Create -parent .$TextFrame -type "Info" -button [list [msgcat::mc "ok"] [msgcat::mc "cancel"]] -title [msgcat::mc "Nothing found - restart?"] -message [msgcat::mc "Nothing found or the end of the document reached!\n\nShould we restart on the begin of the document?"]]
  } else \
  {
   ::FR_messageBox::Create -parent .$TextFrame -type "Info" -title [msgcat::mc "Nothing found"] -message [msgcat::mc "I didn't find: %s" $SearchText]
   set Answer 0
  }

  if {$Answer} \
  {
   .$TextFrame mark set insert 1.0
   focus .$VAR(bottomframe).search_field
   ::tk_edit::FindText $SearchText 1
  } else \
  {
   focus .$TextFrame
  }
 }
}


#---if anything is selected or the cursor is on a word: we copy this in the search field
proc ::tk_edit::FindTextPrepare {{SifCall 0}} \
{
 variable VAR
 
 set Search ""
 set VAR(Search) ""
 
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 
 # if no text frame exists we return
 if {$TextFrame eq ""} {return}
 
 append TextFrame ".t"
 
 #set VAR(TextFrame) $TextFrame
 #set VAR(InsertPosition) [.$TextFrame index insert]
 
 set SelPos [.$TextFrame tag ranges sel]

 if {$SelPos ne ""} \
 {
  set Search [.$TextFrame get [lindex $SelPos 0] [lindex $SelPos 1]]  
 }

 if {$Search eq ""} \
 {
  set Search [.$TextFrame get "insert wordstart" "insert wordend"]
 }

 if {$SifCall == 0} \
 {
   focus .$VAR(bottomframe).search_field
 
   if {$Search ne ""} \
   {
     .$VAR(bottomframe).search_field delete 0 end
     .$VAR(bottomframe).search_field insert end $Search
   }
 }
  
 # we save the wished search phrase for the search in files function
 set VAR(Search) $Search
}
 

#---the replace function
proc ::tk_edit::ReplaceText {Replace} \
{
 variable VAR
 
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 append TextFrame ".t"
 
 if {$Replace eq ""} {return}
 
 set SelPos [.$TextFrame tag ranges sel]

 if {$SelPos ne ""} \
 {
  eval .$TextFrame delete $SelPos 
  set StartPos [lindex $SelPos 0]
  .$TextFrame insert $StartPos $Replace
  .$TextFrame mark set insert "$StartPos + [string length $Replace] c"
 }
 
 .$VAR(bottomframe).replace_button configure -state disable
}


#---the search function in multiple files
proc ::tk_edit::FindInFiles {} \
{
 variable VAR
 
 # first we look if there anything select
 ::tk_edit::FindTextPrepare 1
 
 ::tk_edit::ProcStart ::tk_edit::SearchInFiles search_in_files.tcl
}


#---the selection dialog for the font of the text window
proc ::tk_edit::SelectFont {} \
{
 variable VAR; variable CONFIG
 
 ::tk_edit::Packageload FR_fontSelect
 
 set TextFrame [::FR_notebook::CurrentFramePath .$VAR(nb)]
 append TextFrame ".t"
 
 set ReturnValue [::FR_fontSelect::Create -parent $VAR(Parent) -initialfont [.$TextFrame cget -font]]
 
 if {$ReturnValue eq ""} {return 0}
 
 set CONFIG(TextFont) $ReturnValue
 
 foreach Arg [::FR_notebook::GetAllTabs .$VAR(nb)] \
 {  
  set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $Arg]
  append TextFrame ".t"
  .$TextFrame configure -font $ReturnValue
 }
  
 .$VAR(procframe).procs configure -font $ReturnValue
}


#---This create the menu
proc ::tk_edit::Menu {} \
{
 variable VAR
 
 # the file menu
 menubutton .$VAR(menuframe).file -text [msgcat::mc "File"] -menu .$VAR(menuframe).file.choice -underline 0
 menu .$VAR(menuframe).file.choice -tearoff 0

 .$VAR(menuframe).file.choice add command -label [msgcat::mc "New file"] -accelerator "Ctr+n" -command {::tk_edit::NewFile}
 .$VAR(menuframe).file.choice add separator
 .$VAR(menuframe).file.choice add command -label [msgcat::mc "Open"] -accelerator "Ctr+o" -command {::tk_edit::OpenFile}
 .$VAR(menuframe).file.choice add separator
 .$VAR(menuframe).file.choice add command -label [msgcat::mc "Save"] -accelerator "Ctr+s" -command {::tk_edit::Save}
 .$VAR(menuframe).file.choice add command -label [msgcat::mc "Save as"] -accelerator "Alt+s" -command {::tk_edit::SaveAs}
 .$VAR(menuframe).file.choice add separator
 .$VAR(menuframe).file.choice add command -label [msgcat::mc "Close file"] -accelerator "Ctr+w" -command {::tk_edit::CloseFile}
 .$VAR(menuframe).file.choice add separator
 .$VAR(menuframe).file.choice add command -label [msgcat::mc "Exit"] -accelerator "Ctr+q" -command {::tk_edit::Exit}

 bind .$VAR(menuframe).file.choice <Control-n> [list ::FR_general::MenuBinding .$VAR(menuframe).file.choice 0]
 bind Text <Control-n> {::tk_edit::NewFile}
 bind .$VAR(menuframe).file.choice <Control-o> [list ::FR_general::MenuBinding .$VAR(menuframe).file.choice 2]
 bind Text <Control-o> {::tk_edit::OpenFile}
 bind .$VAR(menuframe).file.choice <Control-s> [list ::FR_general::MenuBinding .$VAR(menuframe).file.choice 4]
 bind Text <Control-s> {::tk_edit::Save}
 bind .$VAR(menuframe).file.choice <Alt-s> [list ::FR_general::MenuBinding .$VAR(menuframe).file.choice 5]
 bind Text <Alt-s> {::tk_edit::SaveAs}
 bind .$VAR(menuframe).file.choice <Control-w> [list ::FR_general::MenuBinding .$VAR(menuframe).file.choice 7]
 bind Text <Control-w> {::tk_edit::CloseFile}
 bind . <Control-q> {::tk_edit::Exit}

 pack .$VAR(menuframe).file -side left

 # edit menu
 menubutton .$VAR(menuframe).edit -text [msgcat::mc "Edit"] -menu .$VAR(menuframe).edit.choice -underline 0
 menu .$VAR(menuframe).edit.choice -tearoff 0

 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Undo"] -accelerator "Ctr+z" -command {::tk_edit::UnDo}
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Redo"] -accelerator "Ctr+y" -command {::tk_edit::ReDo}
 .$VAR(menuframe).edit.choice add separator
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Cut"] -accelerator "Ctr+x" -command {:tk_edit::TextCut}
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Copy"] -accelerator "Ctr+c" -command {::tk_edit::TextCopy}
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Paste"] -accelerator "Ctr+v" -command {::tk_edit::TextPast}
 .$VAR(menuframe).edit.choice add separator
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Select all"] -accelerator "Ctr+a" -command {::tk_edit::SelectAll}
 .$VAR(menuframe).edit.choice add separator
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Refresh highlighting"] -accelerator "Ctr+h" -command {::tk_edit::RefreshHighlighting}
 .$VAR(menuframe).edit.choice add separator
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Comment region"] -accelerator "Ctr+r" -command [list ::tk_edit::Comment 0]
 .$VAR(menuframe).edit.choice add command -label [msgcat::mc "Uncomment region"] -accelerator "Ctr+t" -command [list ::tk_edit::Comment 1]

 bind Text <Control-z> {::tk_edit::UnDo}
 bind .$VAR(menuframe).edit.choice <Control-z> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 0]
 bind Text <Control-y> {::tk_edit::ReDo}
 bind .$VAR(menuframe).edit.choice <Control-y> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 1]
 bind Text <Control-x> {::tk_edit::TextCut; break}
 bind .$VAR(menuframe).edit.choice <Control-x> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 3]
 bind Text <Control-c> {::tk_edit::TextCopy; break}
 bind .$VAR(menuframe).edit.choice <Control-c> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 4]
 bind Text <Control-v> {::tk_edit::TextPast; break}
 bind .$VAR(menuframe).edit.choice <Control-v> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 5]
 bind Text <Control-a> {::tk_edit::SelectAll; break}
 bind .$VAR(menuframe).edit.choice <Control-a> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 7]
 bind Text <Control-h> {::tk_edit::RefreshHighlighting}
 bind .$VAR(menuframe).edit.choice <Control-h> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 9]
 bind Text <Control-r> {list ::tk_edit::Comment 0}
 bind .$VAR(menuframe).edit.choice <Control-r> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 11]
 bind Text <Control-t> {list ::tk_edit::Comment 1}
 bind .$VAR(menuframe).edit.choice <Control-t> [list ::FR_general::MenuBinding .$VAR(menuframe).edit.choice 12]

 pack .$VAR(menuframe).edit -side left

 # config menu
 menubutton .$VAR(menuframe).config -text [msgcat::mc "Configuration"] -menu .$VAR(menuframe).config.choice -underline 0
 menu .$VAR(menuframe).config.choice -tearoff 0

 .$VAR(menuframe).config.choice add checkbutton -label [msgcat::mc "Indent lines"] -accelerator "Ctr+i" -onvalue 1 -offvalue 0 -variable ::tk_edit::CONFIG(TextIndent)
 .$VAR(menuframe).config.choice add separator
 .$VAR(menuframe).config.choice add checkbutton -label [msgcat::mc "Create backup file"] -accelerator "Ctr+b" -onvalue 1 -offvalue 0 -variable ::tk_edit::CONFIG(Backup)
 .$VAR(menuframe).config.choice add separator
 .$VAR(menuframe).config.choice add checkbutton -label [msgcat::mc "Show procedures window"] -accelerator "Ctr+p" -onvalue 1 -offvalue 0 -variable ::tk_edit::CONFIG(ShowProcWin) -command [list ::tk_edit::ProcWindow] 
 .$VAR(menuframe).config.choice add checkbutton -label [msgcat::mc "Sort procs alphabetically"] -accelerator "Ctr+u" -onvalue 1 -offvalue 0 -variable ::tk_edit::CONFIG(SortProcWin) -command [list ::tk_edit::FindProcs] 
 .$VAR(menuframe).config.choice add separator
 .$VAR(menuframe).config.choice add command -label [msgcat::mc "Select textfont"] -accelerator "Ctr+m" -command {::tk_edit::SelectFont}


 bind .$VAR(menuframe).config.choice <Control-i> [list ::FR_general::MenuBinding .$VAR(menuframe).config.choice 0]
 bind .$VAR(menuframe).config.choice <Control-b> [list ::FR_general::MenuBinding .$VAR(menuframe).config.choice 2]
 bind .$VAR(menuframe).config.choice <Control-p> [list ::FR_general::MenuBinding .$VAR(menuframe).config.choice 4]
 bind .$VAR(menuframe).config.choice <Control-u> [list ::FR_general::MenuBinding .$VAR(menuframe).config.choice 5]
 bind .$VAR(menuframe).config.choice <Control-m> [list ::FR_general::MenuBinding .$VAR(menuframe).config.choice 7]
 
 pack .$VAR(menuframe).config -side left

 # help menu
 menubutton .$VAR(menuframe).about -text [msgcat::mc "About"] -menu .$VAR(menuframe).about.choice -underline 0
 menu .$VAR(menuframe).about.choice -tearoff 0

 .$VAR(menuframe).about.choice add command -label [msgcat::mc "About"] -accelerator "Alt+a" -command [list ::FR_messageBox::Create -type About -nspace $VAR(NameSpace)]

 bind .$VAR(menuframe).about.choice <Alt-KeyPress-a> [list ::FR_general::MenuBinding .$VAR(menuframe).about.choice 0]

 pack .$VAR(menuframe).about -side right

 # the menu for the notebook widget 
 menu .$VAR(mainframe).menu_notebook -tearoff 0
 .$VAR(mainframe).menu_notebook add command -label [msgcat::mc "Close"] -command {::tk_edit::CloseFile $::tk_edit::VAR(NbButtonFile)}
}


#---we create all widgets
proc ::tk_edit::Gui {} \
{
 variable VAR; variable CONFIG
 
 # the notebook widget
 set VAR(nb) $VAR(mainframe).notebook
 ::FR_notebook::Create .$VAR(nb) -borderwidth $VAR(BorderWidth) -command ::tk_edit::FindProcs
 pack .$VAR(mainframe).notebook -expand yes -fill both -side left
  
 # the search button
 ::FR_entry::Create .$VAR(bottomframe).search_field
 .$VAR(bottomframe).search_field configure -width 20
 button .$VAR(bottomframe).search_button -text [msgcat::mc "Search"] -padx 10 -command {::tk_edit::FindText [.$::tk_edit::VAR(bottomframe).search_field get]} 
 
 pack .$VAR(bottomframe).search_field -side left -anchor w
 pack .$VAR(bottomframe).search_button -side left -anchor w
 
 bind Text <Control-f> {::tk_edit::FindTextPrepare}
 bind .$VAR(bottomframe).search_field <KeyPress-Return> [list .$VAR(bottomframe).search_button invoke]
 bind . <KeyPress-F3> [list .$VAR(bottomframe).search_button invoke]
 
 # the case sensitive button
 checkbutton .$VAR(bottomframe).casesens -text [msgcat::mc "Case sensitive search"] -justify left -onvalue 1 -offvalue 0 -variable ::tk_edit::CONFIG(CaseSens) 
 pack .$VAR(bottomframe).casesens -side left -anchor w -padx {0 10}
 
 # the replace button
 ::FR_entry::Create .$VAR(bottomframe).replace_field
 .$VAR(bottomframe).replace_field configure -width 20
 button .$VAR(bottomframe).replace_button -text [msgcat::mc "Replace"] -padx 10 -command {::tk_edit::ReplaceText [.$::tk_edit::VAR(bottomframe).replace_field get]} -state disable

 pack .$VAR(bottomframe).replace_field -side left -anchor w
 pack .$VAR(bottomframe).replace_button -side left -anchor w
 
 bind . <Control-r> [list .$VAR(bottomframe).replace_button invoke]
 bind .$VAR(bottomframe).replace_field <KeyPress-Return> [list .$VAR(bottomframe).replace_button invoke]
 
 # the search in files button
 button .$VAR(bottomframe).sif_button -text [msgcat::mc "Search in files"] -padx 10 -command {::tk_edit::FindInFiles} 

 pack .$VAR(bottomframe).sif_button -side right -anchor e
 
 bind Text <Control-F> [list .$VAR(bottomframe).sif_button invoke]
 
 # the proc window
 button .$VAR(procframe).button -text [msgcat::mc "Procedures"] -relief flat -command {::tk_edit::FindProcs}
 text .$VAR(procframe).procs -width 20 -wrap none -bg $CONFIG(TextBackGround) -fg $CONFIG(TextForeGround) -state disable -cursor top_left_arrow -xscrollcommand ".$VAR(procframe).xsbar set" -yscrollcommand ".$VAR(procframe).ysbar set" -font $CONFIG(TextFont)
 scrollbar .$VAR(procframe).xsbar -orient horizontal -command ".$VAR(procframe).procs xview"
 scrollbar .$VAR(procframe).ysbar -orient vertical -command ".$VAR(procframe).procs yview"

 ::tk_edit::ProcWindow

 grid rowconfigure .$VAR(procframe) 1 -weight 1
}


#---pack and unpack routine for the procedure window
proc ::tk_edit::ProcWindow {} \
{
 variable VAR; variable CONFIG
 
 if {$CONFIG(ShowProcWin)} \
 {
  pack .$VAR(procframe) -side right -fill y -padx 5 -pady 5
  grid .$VAR(procframe).button -row 0 -column 0 -sticky ew -padx 5 -columnspan 2 -pady {5 0}
  grid .$VAR(procframe).procs -row 1 -column 0 -sticky nsew -padx {5 0} -pady {5 0}
  grid .$VAR(procframe).ysbar -row 1 -column 1 -sticky ns -padx {0 5} -pady {5 0}
  grid .$VAR(procframe).xsbar -row 2 -column 0 -sticky ew -padx {5 0} -pady {0 5}
 } else \
 {
  # we check if the window is already packed (it's not at the first start)
  if {[winfo manager .$VAR(procframe)] ne ""} \
  {
   # we will forget the text window with its scrollbars...
   foreach Arg [list $VAR(procframe).button $VAR(procframe).procs $VAR(procframe).ysbar $VAR(procframe).xsbar] \
   {
    grid forget .$Arg
   }
   #...and the frame itself
   pack forget .$VAR(procframe)
  }
 }
}


#---we are looking in the current text window for procedures
proc ::tk_edit::FindProcs {} \
{
 variable VAR; variable CONFIG
 
 .$VAR(procframe).procs configure -state normal
 
 .$VAR(procframe).procs delete 1.0 end
 
 set Tags [.$VAR(procframe).procs tag names]

 if {$Tags ne ""} \
 {
  eval .$VAR(procframe).procs tag delete $Tags
 }
 
 set Tab [::FR_notebook::CurrentTab .$VAR(nb)]
 #isn't a file opened we return
 if {$Tab eq ""} {return}
 
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $Tab]
 append TextFrame ".t"
 
 set Text [split [.$TextFrame get 1.0 end-1c] "\n"]
 set TagList ""
 
 foreach Arg $Text \
 { 
  set Result [regexp -all -inline {^([\s\t]*)(proc)+([^\{]*)} $Arg]
  set ResultLength [llength $Result]
  
  if {$ResultLength > 3} \
  {
   for {set i 0} {$i < $ResultLength} {incr i 4} \
   {
     set Tag [lindex $Result [expr {$i+3}]]
     set Tag [string trim $Tag]
     lappend TagList $Tag
     
     set SearchInText($Tag) [lindex $Result $i]
     set SearchInText($Tag) [string trim $SearchInText($Tag)]
    }
  }
 }
 
 if {$CONFIG(SortProcWin)} \
 {
  set TagList [lsort $TagList]
 }
 
 foreach Tag $TagList \
 {
  ::tk_edit::Generate_ProcsTags $Tab $Tag $SearchInText($Tag)
 }
 
 .$VAR(procframe).procs configure -state disable
}


#---we show the procs and generate a link with the search pattern
proc ::tk_edit::Generate_ProcsTags {Tab Tag SearchInText} \
{
 variable VAR
 
 set Link $Tag
 
 # we create the tag
 .$VAR(procframe).procs tag configure $Tag -foreground white -underline no
 .$VAR(procframe).procs tag bind $Tag <Enter> [list .$VAR(procframe).procs tag configure $Tag -underline yes]
 .$VAR(procframe).procs tag bind $Tag <Leave> [list .$VAR(procframe).procs tag configure $Tag -underline no]
 .$VAR(procframe).procs tag bind $Tag <ButtonPress> [list ::tk_edit::SearchProcs $Tab $SearchInText]
 
 # ...and show the link
 .$VAR(procframe).procs insert end $Link $Tag
 .$VAR(procframe).procs insert end "\n"
}


#---we search the proc in the text widget
proc ::tk_edit::SearchProcs {Tab SearchInText} \
{
 variable VAR
 
 # we raise the wished tab in the foreground if it exists
 if {[lsearch [::FR_notebook::GetAllTabs .$VAR(nb)] $Tab] != -1} \
 {
  ::FR_notebook::Raise .$VAR(nb) -tab $Tab
 } else \
 {
  # the file isn't yet opened, we put out a warning and refresh the procs window
  ::FR_messageBox::Create -type "Info" -title [msgcat::mc "File not found"] -message [msgcat::mc "Sorry, the file:\n%s\n isn't opened anymore." $Tab]]
  ::tk_edit::FindProcs
  return
 }
 
 set TextFrame [::FR_notebook::FramePath .$VAR(nb) -tab $Tab]
 append TextFrame ".t"
 
 set Found [.$TextFrame search -count Cnt -- $SearchInText 1.0 end]
 
 # if the proc was deleted in the interval we refresh the procs window and return
 if {$Found eq ""} \
 {
  ::FR_messageBox::Create -type "Info" -title [msgcat::mc "Proc not found"] -message [msgcat::mc "Sorry, we couldn't find:\n%s\nHave you delete this?" $Tab]
  ::tk_edit::FindProcs
  return
 }  
   
 .$TextFrame tag remove sel 1.0 end
 .$TextFrame tag add sel $Found "$Found + $Cnt chars"
 set Pos "$Found wordend"
  
 .$TextFrame mark set insert $Pos
 .$TextFrame see insert
}


#---load the overtaken file or the last file(s) or an empty file after startup
proc ::tk_edit::FirstStart {} \
{
 global argc argv
 variable CONFIG
 
 if {$argc} \
 {
  set File [file normalize $argv]
  
  if {[file exists $File]} \
  {
   ::tk_edit::OpenFile $File
  } else \
  {
   ::tk_edit::NewFile
  }
  
  return 0
 }

 if {$CONFIG(CurrentFiles) ne ""} \
 {
   foreach File $CONFIG(CurrentFiles) \
   {
    if {[file exists $File]} {::tk_edit::OpenFile $File}
   }
 } else \
 {
  ::tk_edit::NewFile
 }
}


::tk_edit::Main